<?php

/**
 * ACL management class
 *
 * LICENSE:
 *
 * Copyright (c) 2012 Diego Zuccato <ndk@csshl.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * o Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * o Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * o The names of the authors may not be used to endorse or promote
 *   products derived from this software without specific prior written
 *   permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * @author      Diego Zuccato <ndk@csshl.net>
 * @copyright   2010 Diego Zuccato
 * @license     http://opensource.org/licenses/bsd-license.php New BSD License
 */
/*
 * From mail sent to the ML:
 * > A good way to represent caps could be a list of comma-separed words, and
 * > a good way to represent an ACL could be a set of 3 (extended: a cap
 * > could be prefixed by a '-' to indicate that it must *not* be set) caps
 * > sets separed by pipes. More formally:
 * > LETNUM := 'a'..'z' | 'A'..'Z' | '0'..'9' | '_'
 * > CAP := LETNUM | LETNUM CAP
 * > CAPS := '' | CAP | CAP ',' CAPS
 * > ECAP := CAP | '-' CAP
 * > ECAPS := '' | ECAP | ECAP ',' ECAPS
 * > ACL := ECAPS '|' ECAPS' '|' ECAPS
 *
 * Extension: caps and ACLs can include levels! cap[=lev][,cap=lev]*
 * A cap w/o level is equal to a cap with level=0.
 * An acl is verified iif:
 * - the checked caps set is a superset of acl's caps
 * - every cap in acl's set have a level <= the corresponding compared cap level
 *
 * For example, given an ACL of "level=2,-disabled":
 * - "level=5,disabled" is rejected ('disabled' is set)
 * - "level=1,admin" is rejected (1 < 2)
 * - "level=2,foo=3,bar=0" is accepted ('foo' and 'bar' aren't in ACL, but it's OK)
 * - "foo=100,bar=abc" is rejected (missing required 'level')
*/

define('ALPHA1', "abcdefghijklmnopqrstuvwxyz");
define('ALPHA2', "ABCDEFGHIJKLMNOPQRSTUVWXYZ");
define('NUMBERS', "0123456789");
define('LETNUM', ALPHA1.ALPHA2.NUMBERS."_");

class ACL {
    // Accept an array (one extended capability per *key* -- value is currently
    // ignored and will be used for the level!) or a string (packed extended
    // capabilities -- duplicated capabilities are overwritten and only the last one remains)
    public function __construct($ecaps) {
	if(is_array($ecaps)) {
	    // Sanitize input array:
	    // - remove illegal keys (not valid caps)
	    // - make sure values are positive integers
	    foreach($ecaps as $k => $v) {
		$t=$this->parse_ecap("$k=$v", TRUE);
		if(FALSE===$t || $t['key']!=$k) { // invalid cap or different key: discard
		    unset($ecaps[$k]);
		} else {
		    $ecaps[$k]=$t['val'];	// sanitize value
		}
	    }
	} else if(is_string($ecaps)) {
	    $ecaps=$this->parse_ecaps_string($ecaps);
	} else {	// Ops! Don't know what to do, so I leave ACL empty
	    trigger_error("Bad ACL constructor parameter -- leaving ACL empty!", E_USER_WARNING);
	    return;
	}
	foreach($ecaps as $cap => $v) {
	    if(substr($cap, 0, 1) == '-') {
		$this->neg_acl[substr($cap, 1)]=$v;
	    } else {
		$this->pos_acl[$cap]=$v;
	    }
	}

	$this->pcaps=count($this->pos_acl); // Useless to count every time it's needed to verify an ACL

	$tst=array_intersect(array_keys($this->pos_acl), array_keys($this->neg_acl));
	if(count($tst)) {
	    trigger_error("Never-satisfiable ACL: pos and neg caps overlap for ". implode(",",$tst), E_USER_WARNING);
	}
    }

    // Gets a capability set and returns if the ACL allows access (TRUE) or not (FALSE)
    public function test($cap) {
	if(is_array($cap)) $caparr=$cap;
	else if(is_string($cap)) $caparr=$this->parse_ecaps_string($cap);
	else return FALSE;

	$pos=count(array_intersect(array_keys($this->pos_acl), array_keys($caparr)));
	$neg=count(array_intersect(array_keys($this->neg_acl), array_keys($caparr)));

	// There must be *NO* negatives and *ALL* positives
	$rv=($neg==0)&&($pos==$this->pcaps);

	if($rv) { // Time to check levels!
	    foreach($this->pos_acl as $cap => $lev) {
		if($caparr[$cap]<$lev) {
		    $rv=FALSE;
		    break; // bail out: level too low!
		}
	    }
	}

	return $rv;
    }

    public function __toString() {
	$tmp=$this->pos_acl;
	foreach($this->neg_acl as $ncap => $nval) {
	    $tmp["-".$ncap]=$nval;
	}
	$out="";
	foreach($tmp as $k => $v) {
	    $out .= ($out==""?"":",")."$k=$v";
	}
	return $out;
    }

    //************************
    private function  parse_ecaps_string($caps, $allow_neg=FALSE) {
	$tcaps = array();
	$tmp=explode(",", $caps);
	foreach($tmp as $tok) { // Probably this can be optimized
	    $t=$this->parse_ecap($tok, $allow_neg);
	    if(is_array($t)) {
		$tcaps[$t['key']] = $t['val'];
	    }
	}
	return $tcaps;
    }

    // Gets a single ecap string and returns an array:
    // 'key' => ecap name
    // 'val' => ecap value
    // Returns FALSE if the capability is invalid
    private function parse_ecap($tok, $allow_neg) {
	$rv=array();
	$start=0;
	if($tok[0]=='-' && $allow_neg) { // *Possibly* a neg cap
	    $start=1; // Skip the initial '-'
	}
	$kl=strspn($tok, LETNUM, $start);
	if($kl) { // at least one valid char as key
	    $kl+=$start;
	    $rv['key'] = substr($tok, 0, $kl);
	    $rv['val'] = 0;
	    if(strlen($tok)>$kl && "="==$tok[$kl]) {
		// value present
		$rv['val'] = (int)substr($tok, $kl+1, strspn($tok, NUMBERS, $kl+1));
	    }
	} else {	// Invalid key ==> invalid cap (no name and val)
	    $rv=FALSE;
	}

	return $rv;
    }

    private $pos_acl=array(), $neg_acl=array(), $pcaps=0;
}
